from datetime import datetime
from typing import Annotated, Optional, Union
from uuid import UUID

import strawberry
from strawberry.federation.schema_directives import Key
from strawberry.unset import UNSET

from oauth2_lib.strawberry import authenticated_field
from orchestrator.db.models import FixedInputTable, SubscriptionTable
from orchestrator.domain.base import SubscriptionModel
from orchestrator.graphql.pagination import EMPTY_PAGE, Connection
from orchestrator.graphql.resolvers.process import resolve_processes
from orchestrator.graphql.schemas.default_customer import DefaultCustomerType
from orchestrator.graphql.schemas.process import ProcessType
from orchestrator.graphql.schemas.product import ProductModelGraphql
from orchestrator.graphql.types import GraphqlFilter, GraphqlSort, OrchestratorInfo
from orchestrator.graphql.utils.get_subscription_product_blocks import (
    ProductBlockInstance,
    get_subscription_product_blocks,
)
from orchestrator.settings import app_settings
from orchestrator.types import SubscriptionLifecycle

federation_key_directives = [Key(fields="subscriptionId", resolvable=UNSET)]


@strawberry.federation.interface(description="Virtual base interface for subscriptions", keys=["subscriptionId"])
class SubscriptionInterface:
    subscription_id: UUID
    customer_id: str
    product: ProductModelGraphql
    description: str
    start_date: Optional[datetime]
    end_date: Optional[datetime]
    status: SubscriptionLifecycle
    insync: bool
    note: Optional[str]

    @strawberry.field(description="Return all products block instances of a subscription")  # type: ignore
    async def product_block_instances(
        self, tags: Optional[list[str]] = None, resource_types: Optional[list[str]] = None
    ) -> list[ProductBlockInstance]:
        return await get_subscription_product_blocks(self.subscription_id, tags, resource_types)

    @strawberry.field(description="Return all products blocks that are part of a subscription", deprecation_reason="changed to product_block_instances")  # type: ignore
    async def product_blocks(
        self, tags: Optional[list[str]] = None, resource_types: Optional[list[str]] = None
    ) -> list[ProductBlockInstance]:
        return await get_subscription_product_blocks(self.subscription_id, tags, resource_types)

    @strawberry.field(description="Return fixed inputs")  # type: ignore
    async def fixed_inputs(self) -> strawberry.scalars.JSON:
        fixed_inputs: list[FixedInputTable] = FixedInputTable.query.filter(
            FixedInputTable.product_id == self.product.product_id  # type: ignore
        ).all()
        return [{"field": fi.name, "value": fi.value} for fi in fixed_inputs]

    @authenticated_field(description="Returns list of processes of the subscription")  # type: ignore
    async def processes(
        self,
        info: OrchestratorInfo,
        filter_by: Union[list[GraphqlFilter], None] = None,
        sort_by: Union[list[GraphqlSort], None] = None,
        first: int = 10,
        after: int = 0,
    ) -> Connection[ProcessType]:
        filter_by_with_related_processes = (filter_by or []) + [
            GraphqlFilter(field="subscriptionId", value=str(self.subscription_id))
        ]
        return await resolve_processes(info, filter_by_with_related_processes, sort_by, first, after)

    @authenticated_field(description="Returns list of subscriptions that use this subscription")  # type: ignore
    async def in_use_by_subscriptions(
        self,
        info: OrchestratorInfo,
        filter_by: Union[list[GraphqlFilter], None] = None,
        sort_by: Union[list[GraphqlSort], None] = None,
        first: int = 10,
        after: int = 0,
    ) -> Connection[Annotated["SubscriptionInterface", strawberry.lazy(".subscription")]]:
        from orchestrator.graphql.resolvers.subscription import resolve_subscriptions
        from orchestrator.services.subscriptions import query_in_use_by_subscriptions

        in_use_by_query = query_in_use_by_subscriptions(self.subscription_id)
        query_results = in_use_by_query.with_entities(SubscriptionTable.subscription_id).all()
        subscription_ids = [str(s.subscription_id) for s in query_results]
        if not subscription_ids:
            return EMPTY_PAGE
        filter_by_with_related_subscriptions = (filter_by or []) + [
            GraphqlFilter(field="subscriptionIds", value=",".join(subscription_ids))
        ]
        return await resolve_subscriptions(info, filter_by_with_related_subscriptions, sort_by, first, after)

    @authenticated_field(description="Returns list of subscriptions that this subscription depends on")  # type: ignore
    async def depends_on_subscriptions(
        self,
        info: OrchestratorInfo,
        filter_by: Union[list[GraphqlFilter], None] = None,
        sort_by: Union[list[GraphqlSort], None] = None,
        first: int = 10,
        after: int = 0,
    ) -> Connection[Annotated["SubscriptionInterface", strawberry.lazy(".subscription")]]:
        from orchestrator.graphql.resolvers.subscription import resolve_subscriptions
        from orchestrator.services.subscriptions import query_depends_on_subscriptions

        depends_on_query = query_depends_on_subscriptions(self.subscription_id)
        query_results = depends_on_query.with_entities(SubscriptionTable.subscription_id).all()
        subscription_ids = [str(s.subscription_id) for s in query_results]
        if not subscription_ids:
            return EMPTY_PAGE
        filter_by_with_related_subscriptions = (filter_by or []) + [
            GraphqlFilter(field="subscriptionIds", value=",".join(subscription_ids))
        ]
        return await resolve_subscriptions(info, filter_by_with_related_subscriptions, sort_by, first, after)

    @strawberry.field(description="Returns customer of a subscription")  # type: ignore
    def customer(self) -> DefaultCustomerType:
        return DefaultCustomerType(
            fullname=app_settings.DEFAULT_CUSTOMER_FULLNAME,
            shortcode=app_settings.DEFAULT_CUSTOMER_SHORTCODE,
            identifier=app_settings.DEFAULT_CUSTOMER_IDENTIFIER,
        )


@strawberry.experimental.pydantic.type(model=SubscriptionModel, all_fields=True, directives=federation_key_directives)
class Subscription(SubscriptionInterface):
    pass



"""Add Subscriptions search view.

Revision ID: e05bb1967eff
Revises: bed6bc0b197a
Create Date: 2023-03-06 10:09:55.675588

"""
from alembic import op
from sqlalchemy import text

# revision identifiers, used by Alembic.
revision = "e05bb1967eff"
down_revision = "bed6bc0b197a"
branch_labels = None
depends_on = None


def upgrade() -> None:
    conn = op.get_bind()
    subscriptions_search_query = """
WITH rt_info AS (SELECT s.subscription_id,
                        concat_ws(', ',
                                  string_agg(rt.resource_type || ':' || siv.value, ', ' ORDER BY rt.resource_type),
                                  string_agg(distinct 'subscription_instance_id' || ':' || si.subscription_instance_id, ', ')
                            ) AS rt_info
                 FROM subscription_instance_values siv
                          JOIN resource_types rt ON siv.resource_type_id = rt.resource_type_id
                          JOIN subscription_instances si ON siv.subscription_instance_id = si.subscription_instance_id
                          JOIN subscriptions s ON si.subscription_id = s.subscription_id
                 GROUP BY s.subscription_id),
     sub_prod_info AS (SELECT s.subscription_id,
                              array_to_string(
                                      ARRAY ['subscription_id:' || s.subscription_id,
                                          'status:' || s.status,
                                          'insync:' || s.insync,
                                          'subscription_description:' || s.description,
                                          'note:' || coalesce(s.note, ''),
                                          'customer_id:' || s.customer_id,
                                          'product_id:' || s.product_id],
                                      ', ') AS sub_info,
                              array_to_string(
                                      ARRAY ['product_name:' || p.name,
                                          'product_description:' || p.description,
                                          'tag:' || p.tag,
                                          'product_type:', p.product_type],
                                      ', ') AS prod_info
                       FROM subscriptions s
                                JOIN products p ON s.product_id = p.product_id),
     fi_info AS (SELECT s.subscription_id,
                        string_agg(fi.name || ':' || fi.value, ', ' ORDER BY fi.name) AS fi_info
                 FROM subscriptions s
                          JOIN products p ON s.product_id = p.product_id
                          JOIN fixed_inputs fi ON p.product_id = fi.product_id
                 GROUP BY s.subscription_id),
     cust_info AS (SELECT s.subscription_id,
                          string_agg('customer_description: ' || scd.description, ', ') AS cust_info
                   FROM subscriptions s
                            JOIN subscription_customer_descriptions scd ON s.subscription_id = scd.subscription_id
                   GROUP BY s.subscription_id)
SELECT s.subscription_id,
       to_tsvector('simple',
                   concat_ws(', ',
                             spi.sub_info,
                             spi.prod_info,
                             fi.fi_info,
                             rti.rt_info,
                             ci.cust_info)
           ) as tsv
FROM subscriptions s
         LEFT JOIN sub_prod_info spi ON s.subscription_id = spi.subscription_id
         LEFT JOIN fi_info fi ON s.subscription_id = fi.subscription_id
         LEFT JOIN rt_info rti ON s.subscription_id = rti.subscription_id
         LEFT JOIN cust_info ci ON s.subscription_id = ci.subscription_id
         """
    subscriptions_search_view_ddl = (
        f"CREATE MATERIALIZED VIEW IF NOT EXISTS subscriptions_search AS {subscriptions_search_query}"
    )

    refresh_subscriptions_search_fn = """
      CREATE OR REPLACE FUNCTION refresh_subscriptions_search_view()
  RETURNS TRIGGER LANGUAGE plpgsql
  AS $$
  BEGIN
  REFRESH MATERIALIZED VIEW CONCURRENTLY subscriptions_search;
  RETURN NULL;
  END $$;
"""

    conn.execute(text(subscriptions_search_view_ddl))
    conn.execute(text(refresh_subscriptions_search_fn))
    conn.execute(text("CREATE INDEX subscriptions_search_tsv_idx ON subscriptions_search USING GIN (tsv);"))
    conn.execute(
        text("CREATE UNIQUE INDEX subscriptions_search_subscription_id_idx ON subscriptions_search (subscription_id);")
    )

    # Refresh the view when dependent tables change
    conn.execute(
        text(
            "CREATE CONSTRAINT TRIGGER fi_refresh_search AFTER UPDATE ON fixed_inputs DEFERRABLE INITIALLY DEFERRED FOR EACH ROW EXECUTE FUNCTION refresh_subscriptions_search_view();"
        )
    )
    conn.execute(
        text(
            "CREATE CONSTRAINT TRIGGER products_refresh_search AFTER UPDATE ON products DEFERRABLE INITIALLY DEFERRED FOR EACH ROW EXECUTE FUNCTION refresh_subscriptions_search_view();"
        )
    )
    conn.execute(
        text(
            "CREATE CONSTRAINT TRIGGER sub_cust_desc_refresh_search AFTER INSERT OR UPDATE OR DELETE ON subscription_customer_descriptions DEFERRABLE INITIALLY DEFERRED FOR EACH ROW EXECUTE FUNCTION refresh_subscriptions_search_view();"
        )
    )
    conn.execute(
        text(
            "CREATE CONSTRAINT TRIGGER siv_refresh_search AFTER INSERT OR UPDATE OR DELETE ON subscription_instance_values DEFERRABLE INITIALLY DEFERRED FOR EACH ROW EXECUTE FUNCTION refresh_subscriptions_search_view();"
        )
    )
    conn.execute(
        text(
            "CREATE CONSTRAINT TRIGGER sub_refresh_search AFTER INSERT OR UPDATE OR DELETE ON subscriptions DEFERRABLE INITIALLY DEFERRED FOR EACH ROW EXECUTE FUNCTION refresh_subscriptions_search_view();"
        )
    )

    # DROP old text-search column and triggers (irreversible)
    old_triggers = [
        ("subscriptions_ins_trigger", "subscriptions"),
        ("subscriptions_upd_trigger", "subscriptions"),
        ("subscriptions_set_tsv_trigger", "subscriptions"),
        ("subscription_customer_descriptions_trigger", "subscription_customer_descriptions"),
        ("subscription_instance_values_trigger", "subscription_instance_values"),
        ("products_trigger", "products"),
        ("fixed_inputs_trigger", "fixed_inputs"),
    ]
    for trigger, table in old_triggers:
        conn.execute(text(f"DROP TRIGGER IF EXISTS {trigger} ON {table};"))

    functions_to_drop = [
        *[trigger for trigger, _ in old_triggers],
        "tsq_parse(regconfig, text), tsq_parse(text), tsq_parse(text, text)",
        "tsq_process_tokens(regconfig, text[]), tsq_process_tokens(text[])",
        "tsq_tokenize",
        "tsq_tokenize_character",
        "tsq_append_current_token",
        "array_nremove",
        "generate_subscription_tsv",
        "parse_websearch(regconfig, text), parse_websearch(text)",
    ]
    for fn in functions_to_drop:
        conn.execute(text(f"DROP FUNCTION IF EXISTS {fn}"))

    conn.execute(text("DROP TYPE IF EXISTS tsq_state;"))
    conn.execute(text("ALTER TABLE subscriptions DROP COLUMN IF EXISTS tsv;"))

    # Fill the materialized view
    conn.execute(text("REFRESH MATERIALIZED VIEW subscriptions_search;"))


def downgrade() -> None:
    conn = op.get_bind()
    triggers_to_drop = [
        ("fi_refresh_search", "fixed_inputs"),
        ("products_refresh_search", "products"),
        ("sub_cust_desc_refresh_search", "subscription_customer_descriptions"),
        ("siv_refresh_search", "subscription_instance_values"),
        ("sub_refresh_search", "subscriptions"),
    ]
    for trigger, table in triggers_to_drop:
        conn.execute(text(f"DROP TRIGGER IF EXISTS {trigger} ON {table};"))

    conn.execute(text("DROP FUNCTION IF EXISTS refresh_subscriptions_search_view();"))
    conn.execute(text("DROP MATERIALIZED VIEW IF EXISTS subscriptions_search CASCADE;"))



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum
from typing import Callable, Iterator, TypeVar

import strawberry
from more_itertools import partition
from pydantic import BaseModel
from sqlalchemy import Column
from sqlalchemy.sql import expression

from orchestrator.api.error_handling import ProblemDetailException
from orchestrator.db.database import SearchQuery
from orchestrator.db.filters import CallableErrorHandler


@strawberry.enum(description="Sort order (ASC or DESC)")
class SortOrder(Enum):
    ASC = "asc"
    DESC = "desc"


class Sort(BaseModel):
    field: str
    order: SortOrder


GenericType = TypeVar("GenericType")
QueryType = SearchQuery
ValidSortFunctionsByColumnType = dict[str, Callable[[QueryType, SortOrder], QueryType]]


def generic_sorts_validate(
    valid_sort_functions_by_column: ValidSortFunctionsByColumnType,
) -> Callable[[list[Sort]], tuple[Iterator[Sort], Iterator[Sort]]]:
    """Create generic validate sort factory that creates a validate function based on the valid sort dict.

    Args:
        valid_sort_functions_by_column: The sort functions per column

    Returns function that takes sort parameters and returns a list of invalid and valid Sort items.
    """

    def validate_sort_items(sort_by: list[Sort]) -> tuple[Iterator[Sort], Iterator[Sort]]:
        def _is_valid_sort(item: Sort) -> bool:
            return item.field in valid_sort_functions_by_column

        return partition(_is_valid_sort, sort_by)

    return validate_sort_items


def generic_apply_sorting(
    valid_sort_functions_by_column: ValidSortFunctionsByColumnType,
) -> Callable[[QueryType, Iterator[Sort], CallableErrorHandler], QueryType]:
    def _apply_sorting(query: QueryType, sort_by: Iterator[Sort], handle_sort_error: CallableErrorHandler) -> QueryType:
        for item in sort_by:
            field = item.field
            sort_fn = valid_sort_functions_by_column[field]
            try:
                query = sort_fn(query, item.order)
            except ProblemDetailException as exception:
                handle_sort_error(
                    exception.detail,
                    field=field,
                    order=item.order,
                )
            except ValueError as exception:
                handle_sort_error(
                    str(exception),
                    field=field,
                    order=item.order,
                )
        return query

    return _apply_sorting


def generic_sort(
    valid_sort_functions_by_column: ValidSortFunctionsByColumnType,
) -> Callable[[QueryType, list[Sort], CallableErrorHandler], QueryType]:
    valid_sort_functions_by_column_KEYS = list(valid_sort_functions_by_column.keys())
    _validate_sorts = generic_sorts_validate(valid_sort_functions_by_column)
    _apply_sorting = generic_apply_sorting(valid_sort_functions_by_column)

    def _sort(
        query: QueryType,
        sort_by: list[Sort],
        handle_sort_error: CallableErrorHandler,
    ) -> QueryType:
        invalid_sort_items, valid_sort_items = _validate_sorts(sort_by)
        if invalid_list := [{"field": item.field, "order": item.order.value.upper()} for item in invalid_sort_items]:
            handle_sort_error(
                "Invalid sort arguments",
                invalid_sorting=invalid_list,
                valid_sort_keys=valid_sort_functions_by_column_KEYS,
            )

        return _apply_sorting(query, valid_sort_items, handle_sort_error)

    return _sort


def generic_column_sort(field: Column) -> Callable[[SearchQuery, SortOrder], SearchQuery]:
    def sort_function(query: SearchQuery, order: SortOrder) -> SearchQuery:
        if order == SortOrder.DESC:
            return query.order_by(expression.desc(field))
        return query.order_by(expression.asc(field))

    return sort_function



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
from typing import Dict, Optional

import typer
import yaml
from jinja2 import Environment, FileSystemLoader

from orchestrator.cli.generator.generator.migration import generate_product_migration
from orchestrator.cli.generator.generator.product import generate_product
from orchestrator.cli.generator.generator.product_block import generate_product_blocks
from orchestrator.cli.generator.generator.unittest import generate_unit_tests
from orchestrator.cli.generator.generator.workflow import generate_workflows

app: typer.Typer = typer.Typer()


def read_config(config_file: str) -> Optional[Dict]:
    try:
        with open(config_file) as stream:
            try:
                return yaml.safe_load(stream)
            except yaml.YAMLError:
                typer.echo("Failed to parse configuration file.")
    except FileNotFoundError:
        typer.echo(f'File "{config_file}" not found')

    return None


def write_file(path: str, content: str, append: bool, force: bool) -> None:
    typer.echo(path)
    try:
        if not force and os.path.exists(path):
            typer.echo(f"Path {path} already exists. Rerun with the --force flag if you want to overwrite")
            return

        mode = "a" if append else "w"
        with open(f"{path}", mode) as writer:
            writer.write(content)
    except FileNotFoundError:
        typer.echo(f"Writing to {path} failed")


def create_context(
    config_file: str, dryrun: bool, force: bool, python_version: str, tdd: Optional[bool] = False
) -> Dict:
    def writer(path: str, content: str, append: bool = False) -> None:
        if dryrun:
            typer.echo(path)
            typer.echo(content)
        else:
            write_file(path, content, append=append, force=force)

    environment = Environment(
        loader=FileSystemLoader(os.path.join(os.path.dirname(__file__), "generator", "templates")), autoescape=True
    )

    config = read_config(config_file)

    return {
        "config": config,
        "environment": environment,
        "python_version": python_version,
        "tdd": tdd,
        "writer": writer,
    }


# Couple of shared parameters since Typer doesn't have an option to do this at the root level
# Discussion: https://github.com/tiangolo/typer/issues/153


ConfigFile = typer.Option(None, "--config-file", "-cf", help="The configuration file")
DryRun = typer.Option(True, help="Dry run")
TestDrivenDevelopment = typer.Option(True, "--tdd", help="Force test driven development with failing asserts")
Force = typer.Option(False, "--force", "-f", help="Force overwrite of existing files")
PythonVersion = typer.Option("3.9", "--python-version", "-p", help="Python version for generated code")


@app.command(help="Create product from configuration file")
def product(
    config_file: str = ConfigFile,
    dryrun: bool = DryRun,
    force: bool = Force,
    python_version: str = PythonVersion,
) -> None:
    context = create_context(config_file, dryrun=dryrun, force=force, python_version=python_version)

    generate_product(context)


@app.command(help="Create product blocks from configuration file")
def product_blocks(
    config_file: str = ConfigFile,
    dryrun: bool = DryRun,
    force: bool = Force,
    python_version: str = PythonVersion,
) -> None:
    context = create_context(config_file, dryrun=dryrun, force=force, python_version=python_version)

    generate_product_blocks(context)


@app.command(help="Create workflows from configuration file")
def workflows(
    config_file: str = ConfigFile,
    dryrun: bool = DryRun,
    force: bool = Force,
    python_version: str = PythonVersion,
    tdd: bool = TestDrivenDevelopment,
) -> None:
    context = create_context(config_file, dryrun=dryrun, force=force, python_version=python_version, tdd=tdd)

    generate_workflows(context)


@app.command(help="Create unit tests from configuration file")
def unit_tests(
    config_file: str = ConfigFile,
    dryrun: bool = DryRun,
    force: bool = Force,
    python_version: str = PythonVersion,
    tdd: bool = TestDrivenDevelopment,
) -> None:
    context = create_context(config_file, dryrun=dryrun, force=force, python_version=python_version, tdd=tdd)

    generate_unit_tests(context)


@app.command(help="Create migration from configuration file")
def migration(
    config_file: str = ConfigFile,
    dryrun: bool = DryRun,
    force: bool = Force,
    python_version: str = PythonVersion,
) -> None:
    context = create_context(config_file, dryrun=dryrun, force=force, python_version=python_version)

    generate_product_migration(context)
